# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING, Any

import cohere
from cohere.types import UserChatMessageV2

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.common.compat.sdk import BaseHook

if TYPE_CHECKING:
    from cohere.core.request_options import RequestOptions
    from cohere.types import ChatMessages


logger = logging.getLogger(__name__)


class CohereHook(BaseHook):
    """
    Use Cohere Python SDK to interact with Cohere platform using API v2.

    .. seealso:: https://docs.cohere.com/docs

    :param conn_id: :ref:`Cohere connection id <howto/connection:cohere>`
    :param timeout: Request timeout in seconds. Optional.
    :param max_retries: Maximal number of retries for requests. Deprecated, use request_options instead. Optional.
    :param request_options: Dictionary for function-specific request configuration. Optional.
    """

    conn_name_attr = "conn_id"
    default_conn_name = "cohere_default"
    conn_type = "cohere"
    hook_name = "Cohere"

    def __init__(
        self,
        conn_id: str = default_conn_name,
        timeout: int | None = None,
        max_retries: int | None = None,
        request_options: RequestOptions | None = None,
    ) -> None:
        super().__init__()
        self.conn_id = conn_id
        self.timeout = timeout
        self.max_retries = max_retries
        self.request_options = request_options
        self._client: cohere.ClientV2 | None = None

        if self.max_retries:
            warnings.warn(
                "Argument `max_retries` is deprecated. Use `request_options` dict for function-specific request configuration.",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
            if self.request_options is None:
                self.request_options = {"max_retries": self.max_retries}
            else:
                self.request_options.update({"max_retries": self.max_retries})

    def get_conn(self) -> cohere.ClientV2:
        """Return a new or cached Cohere client instance."""
        if self._client is None:
            # create a new client instance if there is no existing client
            conn = self.get_connection(self.conn_id)
            self._client = cohere.ClientV2(
                api_key=conn.password,
                timeout=self.timeout,
                base_url=conn.host or None,
            )
        return self._client

    def create_embeddings(
        self, texts: list[str], model: str = "embed-multilingual-v3.0"
    ) -> list[list[float]]:
        logger.info("Creating embeddings with model: embed-multilingual-v3.0")
        response = self.get_conn().embed(
            texts=texts,
            model=model,
            input_type="search_document",
            embedding_types=["float"],
            request_options=self.request_options,
        )
        # NOTE: Return type `EmbedByTypeResponseEmbeddings` was removed temporarily due to limitations
        # in XCom serialization/deserialization of complex types like Cohere embeddings and Pydantic models.
        #
        # Tracking issue: https://github.com/apache/airflow/issues/50867
        # Once that issue is resolved, XCom (de)serialization of such types will be supported and
        # we can safely restore the `EmbedByTypeResponseEmbeddings` return type here.
        if response.embeddings.float_ is None:
            raise ValueError("Embeddings response is missing float_ field")
        return response.embeddings.float_

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        return {
            "hidden_fields": ["schema", "login", "port", "extra"],
            "relabeling": {
                "password": "API Key",
            },
        }

    def test_connection(
        self,
        model: str = "command-r-plus-08-2024",
        messages: ChatMessages | None = None,
    ) -> tuple[bool, str]:
        try:
            if messages is None:
                messages = [UserChatMessageV2(role="user", content="hello world!")]
            self.get_conn().chat(model=model, messages=messages)
            return True, "Connection successfully established."
        except Exception as e:
            return False, f"Unexpected error: {str(e)}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from airflow.providers.cohere.hooks.cohere import CohereHook
from airflow.providers.common.compat.sdk import BaseOperator

if TYPE_CHECKING:
    from cohere.core.request_options import RequestOptions

    from airflow.providers.common.compat.sdk import Context


class CohereEmbeddingOperator(BaseOperator):
    """
    Creates the embedding base by interacting with cohere hosted services.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CohereEmbeddingOperator`

    :param input_text: single string text or list of text items that need to be embedded.
    :param conn_id: Optional. The name of the Airflow connection to get connection
        information for Cohere. Defaults to "cohere_default".
    :param timeout: Timeout in seconds for Cohere API.
    :param max_retries: Number of times to retry before failing.
    :param request_options: Request-specific configuration.
        Fields:
        - timeout_in_seconds: int. The number of seconds to await an API call before timing out.

        - max_retries: int. The max number of retries to attempt if the API call fails.

        - additional_headers: typing.Dict[str, typing.Any]. A dictionary containing additional parameters to spread into the request's header dict

        - additional_query_parameters: typing.Dict[str, typing.Any]. A dictionary containing additional parameters to spread into the request's query parameters dict

        - additional_body_parameters: typing.Dict[str, typing.Any]. A dictionary containing additional parameters to spread into the request's body parameters dict
    """

    template_fields: Sequence[str] = ("input_text",)

    def __init__(
        self,
        input_text: list[str] | str,
        conn_id: str = CohereHook.default_conn_name,
        timeout: int | None = None,
        max_retries: int | None = None,
        request_options: RequestOptions | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        if isinstance(input_text, str):
            input_text = [input_text]
        self.conn_id = conn_id
        self.input_text = input_text
        self.timeout = timeout
        self.max_retries = max_retries
        self.request_options = request_options

    @cached_property
    def hook(self) -> CohereHook:
        """Return an instance of the CohereHook."""
        return CohereHook(
            conn_id=self.conn_id,
            timeout=self.timeout,
            max_retries=self.max_retries,
            request_options=self.request_options,
        )

    def execute(self, context: Context) -> list[list[float]]:
        """Embed texts using Cohere embed services."""
        embedding_response = self.hook.create_embeddings(self.input_text)
        # NOTE: Return type `EmbedByTypeResponseEmbeddings` was removed temporarily due to limitations
        # in XCom serialization/deserialization of complex types like Cohere embeddings and Pydantic models.
        #
        # Tracking issue: https://github.com/apache/airflow/issues/50867
        # Once that issue is resolved, XCom (de)serialization of such types will be supported, and
        # we can safely restore the `EmbedByTypeResponseEmbeddings` return type here.
        return embedding_response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import patch

from airflow.models import Connection
from airflow.providers.cohere.hooks.cohere import (
    CohereHook,
)


class TestCohereHook:
    """
    Test for CohereHook
    """

    def test__get_api_key(self):
        api_key = "test"
        base_url = "http://some_host.com"
        timeout = 150
        with (
            patch.object(
                CohereHook,
                "get_connection",
                return_value=Connection(conn_type="cohere", password=api_key, host=base_url),
            ),
            patch("cohere.ClientV2") as client,
        ):
            hook = CohereHook(timeout=timeout)
            _ = hook.get_conn()
            client.assert_called_once_with(api_key=api_key, timeout=timeout, base_url=base_url)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

from airflow.models import Connection
from airflow.providers.cohere.operators.embedding import CohereEmbeddingOperator


@patch("airflow.providers.cohere.hooks.cohere.CohereHook.get_connection")
@patch("cohere.ClientV2")
def test_cohere_embedding_operator(cohere_client, get_connection):
    """
    Test Cohere client is getting called with the correct key and that
    the execute method returns expected response.
    """
    embedded_obj = [[1.0, 2.0, 3.0]]

    mock_response = MagicMock()
    mock_response.embeddings.float_ = embedded_obj

    api_key = "test"
    base_url = "http://some_host.com"
    timeout = 150
    texts = ["On Kernel-Target Alignment. We describe a family of global optimization procedures"]
    request_options = None

    get_connection.return_value = Connection(conn_type="cohere", password=api_key, host=base_url)
    client_obj = MagicMock()
    cohere_client.return_value = client_obj
    client_obj.embed.return_value = mock_response

    op = CohereEmbeddingOperator(
        task_id="embed",
        conn_id="some_conn",
        input_text=texts,
        timeout=timeout,
        request_options=request_options,
    )

    val = op.execute(context={})
    cohere_client.assert_called_once_with(api_key=api_key, base_url=base_url, timeout=timeout)
    assert val == embedded_obj



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow import DAG
from airflow.providers.cohere.operators.embedding import CohereEmbeddingOperator

with DAG("example_cohere_embedding", schedule=None, start_date=datetime(2023, 1, 1), catchup=False) as dag:
    # [START howto_operator_cohere_embedding]
    texts = [
        "On Kernel-Target Alignment. We describe a family of global optimization procedures",
        " that automatically decompose optimization problems into smaller loosely coupled",
        " problems, then combine the solutions of these with message passing algorithms.",
    ]

    CohereEmbeddingOperator(input_text=texts, task_id="embedding_via_text")
    CohereEmbeddingOperator(input_text=texts[0], task_id="embedding_via_task")
    # [END howto_operator_cohere_embedding]


from tests_common.test_utils.system_tests import get_test_run

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
